lm.groupLasso <- function(XtX, Xty, group, tpk, lambda, beta, mu, eps, max.it){

  env <- get("scoop.par", envir=scoop.env)  
  
  ## INITIALIZATION
  p  <- ncol(XtX)
  pk <- rep.int(tpk,tpk)
  active <- group.norm.rep(beta, tpk) > 0
  beta[!active] <- 0
  it <- 0

  ## Check if optimality is reached at starting point
  dL.min <- env$dL(beta,XtX,Xty,lambda,group,pk,tpk)
  n2.dL  <- group.norm(dL.min,tpk)
  if (max(n2.dL) < eps*p) {
    return(list(beta=beta,status="converged"))
  } else {
    ## UPDATE THE ACTIVE SET / ACTIVE GROUP
    new <- which.max(n2.dL)
    active[group == new] <- TRUE
  }
  while (1) {
    
    it <- it+1
    
    ## (1) OPTIMIZATION OVER A
    ## _____________________________________________________________
    res <- optim(as.matrix(beta[active]), method="L-BFGS-B", fn=env$L, gr=env$dL,
                 XtX=XtX[active,active], Xty=Xty[active],
                 lambda=lambda, group=group[active],pk=pk[active],
                 tpk=tabulate(factor(group[active])),
                 control=list(factr=eps,pgtol=eps))
    beta[active] <- res$par
    beta[active & abs(beta) < eps] <- 0

    ## (2) TEST LAGRAGIAN VIOLATION AND GROUP DELETION IF APPLICABLE
    ## _____________________________________________________________
    dL.min  <- env$dL(beta,XtX,Xty,lambda,group,pk,tpk)
    n2.dL   <- group.norm(dL.min,tpk)
    n2.beta <- group.norm(beta,tpk)
    for (i in intersect(which(n2.beta == 0),group[active])) {
      if (n2.dL[i] == 0) {
        active[group == i] <- FALSE
      }
      beta[group == i] <- 0
    }

    ## (3) OPTIMALITY TESTING AND GROUP ACTIVATION IF APPLICABLE
    ## _____________________________________________________________
    if (it == max.it) {
      return(list(beta=beta,status="max # of iterates reached",
                  max.n2dL=max(n2.dL)))
    }
    if (max(n2.dL) < eps*p) {
      return(list(beta=beta,status="converged"))
    } else {
      ## UPDATE THE ACTIVE SET / ACTIVE GROUP
      new <- which.max(n2.dL)
      active[group == new] <- TRUE
    }
  }
}

L.group.gaussian <- function(beta,XtX,Xty,lambda,group,pk,tpk) {

  lambda <- lambda*sqrt(tpk)
  n2 <- group.norm(beta,tpk)
  
  L <- .5*crossprod(beta,crossprod(XtX,beta))-crossprod(beta,Xty)+sum(lambda*n2)
  
  return(L)
}

dL.group.gaussian <- function(beta,XtX,Xty,lambda,group,pk,tpk) {

  Lambda <- lambda*sqrt(pk)
  n2 <- group.norm.rep(beta,tpk)
  zero <- n2 <= 0

  ## terms whose norm2 is different from zero
  dL <- rep(0,length(beta))

  nabla.f <- drop(crossprod(XtX,beta)) - Xty
  dL[!zero] <- nabla.f[!zero] + Lambda[!zero] * beta[!zero] / n2[!zero]

  ## terms whose norm2 is zero
  n2.nabla.f <- group.norm.rep(nabla.f,tpk)
  to.zero    <- n2.nabla.f[zero] <= Lambda[zero]

  ## which ones can be null ?
  dL[zero][to.zero] <- 0
  dL[zero][!to.zero] <- nabla.f[zero][!to.zero] * (1 - Lambda[zero][!to.zero]/n2.nabla.f[zero][!to.zero])

  return(dL)
}  

lm.coopLasso <- function(XtX, Xty, group, tpk, lambda, beta, mu, eps, max.it){

  env <- get("scoop.par", envir=scoop.env)  
  
  ## INITIALIZATION
  p <- ncol(XtX)
  max.dL.min.old <- Inf
  active.pos <- group.norm.rep(pmax(0, beta),tpk) > 0
  active.neg <- group.norm.rep(pmax(0,-beta),tpk) > 0
  beta[beta > 0 & !active.pos] <- 0
  beta[beta < 0 & !active.neg] <- 0
  it <- 0

  ## Save some time
  pk  <- rep.int(tpk,tpk)
  
  ## Check if optimality is reached at the starting point
  dL.min <- env$dL(beta,XtX,Xty,lambda,group,pk, tpk) 
  n2.dL.pos  <- group.norm(pmax.int(0, dL.min),tpk)
  n2.dL.neg  <- group.norm(pmax.int(0,-dL.min),tpk)
  pmax.n2.dL <- pmax.int(n2.dL.pos,n2.dL.neg)
  if (max(pmax.n2.dL)<eps*p) {
    return(list(beta=beta,status="converged"))
  } else {
    ## UPDATE THE ACTIVE SET / ACTIVE GROUP
    new <- which.max(pmax.n2.dL)
    if (n2.dL.pos[new] > n2.dL.neg[new]) {
      active.neg[group == new] <- TRUE
    } else {
      active.pos[group == new] <- TRUE
    }
  }
  
  while (1) {
    
    it <- it+1
    
    ## (1) OPTIMIZATION OVER A
    ## _____________________________________________________________
    lower.bound <- rep(-Inf,p)
    upper.bound <- rep(+Inf,p)
    lower.bound[active.pos & !active.neg] <- 0
    upper.bound[active.neg & !active.pos] <- 0
    active <- active.pos | active.neg
    res <- optim(beta[active], method="L-BFGS-B", fn=env$L, gr=env$dL, 
                 XtX=XtX[active,active], Xty=Xty[active], lambda=lambda,
                 group=group[active], pk = pk[active],
                 tpk = tabulate(factor(group[active])),
                 lower = lower.bound[active],upper=upper.bound[active],
                 control=list(factr=eps,pgtol=eps))
    beta[active] <- res$par
    beta[active & abs(beta) < eps] <- 0
    
    ## (2) TEST LAGRAGIAN VIOLATION AND GROUP DELETION IF APPLICABLE
    ## _____________________________________________________________
    n2.beta.pos <- group.norm(pmax.int(0,beta),tpk)
    for (i in intersect(which(n2.beta.pos == 0),group[active.pos])){
      beta.tmp <- beta
      beta.tmp[group == i & beta > 0] <- 0
      dL.min <- env$dL(beta.tmp,XtX,Xty,lambda,group,pk,tpk)
      n2.dL.neg <- group.norm(pmax.int(eps,-dL.min), tpk)
      if (n2.dL.neg[i] == 0) {
         active.pos[group == i] <- FALSE
        beta[beta > 0 & group == i] <- 0
      }
    }

    n2.beta.neg <- group.norm(pmax.int(0,-beta),tpk)
    for (i in intersect(which(n2.beta.neg == 0),group[active.neg])){
      beta.tmp <- beta
      beta.tmp[group == i & beta < 0] <- 0
      dL.min <- env$dL(beta.tmp,XtX,Xty,lambda,group,pk,tpk)
      n2.dL.pos <- group.norm(pmax.int(eps, dL.min), tpk)
      if (n2.dL.pos[i] == 0) {
        active.neg[group == i] <- FALSE
        beta[beta < 0 & group == i] <- 0
      }
    }
    
    ## (3) OPTIMALITY TESTING AND GROUP ACTIVATION IF APPLICABLE
    ## _____________________________________________________________
    ## Get the entry for which the constraint is the most violated
    dL.min  <- env$dL(beta,XtX,Xty,lambda,group,pk,tpk)
    max.dL.min <- max(abs(dL.min))
    n2.dL.pos  <- group.norm(pmax.int(0, dL.min),tpk)
    n2.dL.neg  <- group.norm(pmax.int(0,-dL.min),tpk)
    pmax.n2.dL <- pmax.int(n2.dL.pos,n2.dL.neg)
    if (it == max.it) {
      return(list(beta=beta,status="max # of iterates reached",
                  max.n2dL=max(pmax.n2.dL)))
    }
    if (abs(max.dL.min - max.dL.min.old) < eps) {
      return(list(beta=beta,status="no further improvement...",
                  max.n2dL=max.dL.min))
    }
    if (max(pmax.n2.dL)<eps*p) {
      return(list(beta=beta,status="converged"))
    } else {
      ## UPDATE THE ACTIVE SET / ACTIVE GROUP
      max.dL.min.old <- max.dL.min
      new <- which.max(pmax.n2.dL)
      if (n2.dL.pos[new] > n2.dL.neg[new]) {
        active.neg[group == new] <- TRUE
      } else {
        active.pos[group == new] <- TRUE
      }
    }
  }
}

L.coop.gaussian <- function(beta,XtX,Xty,lambda,group,pk,tpk) {

  # Lambda <- Lambda * sqrt(pk) /pk
  lambda <- lambda / sqrt(pk)
  ncoop <- group.norm.rep(pmax.int(0,beta),tpk)+group.norm.rep(pmax.int(0,-beta),tpk)
  
  L <- .5*crossprod(beta,crossprod(XtX,beta))-crossprod(beta,Xty) + sum(lambda*ncoop)
  return(L)
}

dL.coop.gaussian <- function(beta,XtX,Xty,lambda,group,pk,tpk) {

  lambda <- lambda * sqrt(pk)
  a.pos <- group.norm.rep(pmax.int(0, beta),tpk)
  a.neg <- group.norm.rep(pmax.int(0,-beta),tpk)
  
  ## manage cases where alpha+ > 0 or alpha- < 0
  dL <- drop(crossprod(XtX,beta)) - Xty
  neg <- beta < 0
  pos <- beta > 0
  dL[pos] <- dL[pos] + lambda[pos]*pmax.int(rep(0,sum(pos)), beta[pos]) / a.pos[pos]
  dL[neg] <- dL[neg] - lambda[neg]*pmax.int(rep(0,sum(neg)),-beta[neg]) / a.neg[neg]
  
  ## manage cases where beta = 0 and (alpha- = 0 and/or alpha+ = 0)
  n2.theta.pos <- group.norm.rep(pmax.int(0, dL),tpk)
  n2.theta.neg <- group.norm.rep(pmax.int(0,-dL),tpk)
  
  zero.pos  <- abs(beta) == 0 & dL >= 0 & a.neg == 0
  zero.neg  <- abs(beta) == 0 & dL <= 0 & a.pos == 0
  
  pos.to.zero  <- n2.theta.pos[zero.pos] <= lambda[zero.pos]
  neg.to.zero  <- n2.theta.neg[zero.neg] <= lambda[zero.neg]

  dL[zero.pos][!pos.to.zero] <- dL[zero.pos][!pos.to.zero] * (1-lambda[zero.pos][!pos.to.zero]/n2.theta.pos[zero.pos][!pos.to.zero])
  dL[zero.neg][!neg.to.zero] <- dL[zero.neg][!neg.to.zero] * (1-lambda[zero.neg][!neg.to.zero]/n2.theta.neg[zero.neg][!neg.to.zero])
  
  dL[zero.pos][pos.to.zero] <- 0
  dL[zero.neg][neg.to.zero] <- 0
  
  return(dL)
}

