lrm.groupLasso <- function(X, y, group, tpk, lambda, beta, mu, eps, max.it){
  
  env <- get("scoop.par", envir=scoop.env)  
  
  ## INITIALIZATION
  p  <- ncol(X)
  pk <- rep.int(tpk,tpk)
  active <- group.norm.rep(beta, tpk) > 0
  beta[!active] <- 0
  it <- 0

  ## Check if optimality is reached at starting point
  dL.min <- env$dL(c(mu,beta),X,y,lambda,group,pk,tpk)
  if (is.null(mu)) {
    n2.dL  <- group.norm(dL.min,tpk)
  } else {
    n2.dL  <- group.norm(dL.min[-1],tpk)
  }
  if (max(abs(dL.min)) < eps*p) {
    return(list(beta=beta,mu=mu,status="converged"))
  } else {
    ## UPDATE THE ACTIVE SET / ACTIVE GROUP
    new <- which.max(n2.dL)
    active[group == new] <- TRUE
  }
  while (1) {
    
    it <- it+1

    ## (1) OPTIMIZATION OVER A
    ## _____________________________________________________________
    res <- optim(c(mu,beta[active]), method="L-BFGS-B", fn=env$L, gr=env$dL,
                 X=X[,active], y=y,
                 lambda=lambda, group=group[active], pk=pk[active], 
                 tpk=tabulate(factor(group[active])),
                 control=list(factr=eps,pgtol=eps))
    if (is.null(mu)) {
      beta[active] <- res$par
    } else {
      mu <- res$par[1]
      beta[active] <- res$par[-1]
    }
    beta[active & abs(beta)<eps] <- 0
    
    ## (2) TEST LAGRAGIAN VIOLATION AND GROUP DELETION IF APPLICABLE
    ## _____________________________________________________________
    dL.min  <- env$dL(c(mu,beta),X,y,lambda,group,pk,tpk)
    if (is.null(mu)) {
      n2.dL  <- group.norm(dL.min,tpk)
    } else {
      n2.dL  <- group.norm(dL.min[-1],tpk)
    }
    n2.beta <- group.norm(beta,tpk)
    for (i in intersect(which(n2.beta == 0),group[active])) {
      if (n2.dL[i] == 0) {
        active[group == i] <- FALSE
      }
      beta[group == i] <- 0
    }
    
    ## (3) OPTIMALITY TESTING AND GROUP ACTIVATION IF APPLICABLE
    ## _____________________________________________________________
    if (it == max.it) {
      return(list(beta=beta,mu=mu,
                  status="max # of iterates reached",max.n2dL=max(abs(dL.min))))
    }
    if (max(abs(dL.min) < eps*p)) {
      return(list(beta=beta,mu=mu,status="converged"))
    } else {
      ## UPDATE THE ACTIVE SET / ACTIVE GROUP
      new <- which.max(n2.dL)
      active[group == new] <- TRUE
    }
  }
}

L.group.binomial <- function(beta,X,y,lambda,group,pk,tpk) {

  ## Managing intercept
  lambda <- lambda*sqrt(tpk)
  if (length(beta) == length(group)) {
    eta <- X %*% beta
    n2 <- group.norm(beta,tpk)
  } else {
    eta <- cbind(1,X) %*% beta
    n2 <- group.norm(beta[-1],tpk)
  }
  
  L  <- - sum( y * eta - log(1 + exp(eta)) ) + sum(lambda*n2)

  return(L)
}

dL.group.binomial <- function(beta,X,y,lambda,group,pk,tpk) {

  lambda <- lambda * sqrt(pk)
  ## Managing intercept
  if (length(beta) == length(group)) {
    eta <- X %*% beta
    nabla.f <- - t(X) %*% (y-1/(1+exp(-eta)))
    n2 <- group.norm.rep(beta,tpk)
  } else {
    eta <- cbind(1,X) %*% beta
    nabla.f <- - ( t(cbind(1,X)) %*% (y-1/(1+exp(-eta))) )
    lambda <- c(0,lambda)
    n2 <- c(abs(beta[1]),group.norm.rep(beta[-1],tpk))
  }

  zero <- n2 == 0
  ## terms whose norm2 is different from zero
  dL <- rep(0,length(beta))
  dL[!zero] <- nabla.f[!zero] + lambda[!zero] * beta[!zero] / n2[!zero]
  
  ## terms whose norm2 is zero
  if (length(beta) == length(group)) {
    n2.nabla.f <- group.norm.rep(nabla.f,tpk)
  } else {
    n2.nabla.f <- c(abs(nabla.f[1]),group.norm.rep(nabla.f[-1],tpk))
  }
  to.zero    <- n2.nabla.f[zero] <= lambda[zero]
  
  ## which ones can be null ?
  dL[zero][to.zero] <- 0
  dL[zero][!to.zero] <- nabla.f[zero][!to.zero] * (1 - lambda[zero][!to.zero]/n2.nabla.f[zero][!to.zero])

  return(dL)
}  

lrm.coopLasso <- function(X, y, group, tpk, lambda, beta, mu, eps, max.it){

  env <- get("scoop.par", envir=scoop.env)  
  
  ## INITIALIZATION
  max.dL.min.old <- Inf
  p  <- ncol(X)
  pk <- rep.int(tpk,tpk)
  active.pos <- group.norm.rep(pmax.int(0, beta),tpk) > 0
  active.neg <- group.norm.rep(pmax.int(0,-beta),tpk) > 0
  beta[beta > 0 & !active.pos] <- 0
  beta[beta < 0 & !active.neg] <- 0
  it <- 0

  ## Check if optimality is reached at the starting point
  dL.min <- env$dL(c(mu,beta),X,y,lambda,group,pk,tpk)
  if (max(abs(dL.min)) < eps*p) {
    return(list(beta=beta,mu=mu,status="converged"))
  } else {
    ## UPDATE THE ACTIVE SET / ACTIVE GROUP
    if (is.null(mu)) {
      n2.dL.pos <- group.norm(pmax.int(0, dL.min),tpk)
      n2.dL.neg <- group.norm(pmax.int(0,-dL.min),tpk)
    } else {
      n2.dL.pos <- group.norm(pmax.int(0, dL.min[-1]),tpk)
      n2.dL.neg <- group.norm(pmax.int(0,-dL.min[-1]),tpk)
    }
    new <- which.max(pmax.int(n2.dL.pos,n2.dL.neg))
    if (n2.dL.pos[new] > n2.dL.neg[new]) {
      active.neg[group == new] <- TRUE
    } else {
      active.pos[group == new] <- TRUE
    }
  }
  
  while (1) {
    
    it <- it+1
    
    ## (1) OPTIMIZATION OVER A
    ## _____________________________________________________________
    lower.bound <- rep(-Inf,p)
    upper.bound <- rep(+Inf,p)
    lower.bound[active.pos & !active.neg] <- 0
    upper.bound[active.neg & !active.pos] <- 0
    active <- active.pos | active.neg
    res <- optim(c(mu,beta[active]), method="L-BFGS-B", fn=env$L, gr=env$dL, 
                 X=X[,active], y=y, lambda=lambda,
                 group=group[active], pk=pk[active],
                 tpk = tabulate(factor(group[active])),
                 lower = c(-Inf,lower.bound[active]),
                 upper=c(Inf,upper.bound[active]),
                 control=list(factr=eps*p, pgtol=eps*p)
                 )
    if (is.null(mu)) {
      beta[active] <- res$par
    } else {
      mu <- res$par[1]
      beta[active] <- res$par[-1]
    }
    beta[active & abs(beta)<eps] <- 0

    ## (2) TEST LAGRAGIAN VIOLATION AND GROUP DELETION IF APPLICABLE
    ## _____________________________________________________________
    n2.beta.pos <- group.norm(pmax.int(0,beta),tpk)
    for (i in intersect(which(n2.beta.pos == 0),group[active.pos])){
      beta.tmp <- beta
      beta.tmp[group == i & beta > 0] <- 0
      dL.min <- env$dL(c(mu,beta.tmp),X,y,lambda,group,pk,tpk)
      if (is.null(mu)) {
        n2.dL.neg <- group.norm(pmax.int(0,-dL.min),tpk)
      } else {
        n2.dL.neg <- group.norm(pmax.int(0,-dL.min[-1]),tpk)
      }
      if (n2.dL.neg[i] == 0) {
        active.pos[group == i] <- FALSE
        beta[beta > 0 & group == i] <- 0
      }
    }
    
    n2.beta.neg <- group.norm(pmax.int(0,-beta),tpk)
    for (i in intersect(which(n2.beta.neg == 0),group[active.neg])){
      beta.tmp <- beta
      beta.tmp[group == i & beta < 0] <- 0
      dL.min <- env$dL(c(mu,beta.tmp),X,y,lambda,group,pk,tpk)
      if (is.null(mu)) {
        n2.dL.pos <- group.norm(pmax.int(0, dL.min),tpk)
      } else {
        n2.dL.pos <- group.norm(pmax.int(0, dL.min[-1]),tpk)
      }
      if (n2.dL.pos[i] == 0) {
        active.neg[group == i] <- FALSE
        beta[beta < 0 & group == i] <- 0
      }
    }
    
    ## (3) OPTIMALITY TESTING AND GROUP ACTIVATION IF APPLICABLE
    ## _____________________________________________________________
    ## Get the entry for which the constraint is the most violated
    
    dL.min <- env$dL(c(mu,beta),X,y,lambda,group,pk,tpk)
    max.dL.min <- max(abs(dL.min))
    if (it == max.it) {
      return(list(beta=beta,mu=mu,status="max # of iterates reached",
                  max.n2dL=max(abs(dL.min))))
    }
    if (abs(max.dL.min - max.dL.min.old) < eps) {
      return(list(beta=beta,mu=mu,status="no further improvement...",
                  max.n2dL=max.dL.min))
    }
    if (max(abs(dL.min)) < eps*p) {
      return(list(beta=beta,mu=mu,status="converged"))
    } else {
      max.dL.min.old <- max.dL.min
      ## UPDATE THE ACTIVE SET / ACTIVE GROUP
      if (is.null(mu)) {
        n2.dL.pos <- group.norm(pmax.int(0, dL.min),tpk)
        n2.dL.neg <- group.norm(pmax.int(0,-dL.min),tpk)
      } else {
        n2.dL.pos <- group.norm(pmax.int(0, dL.min[-1]),tpk)
        n2.dL.neg <- group.norm(pmax.int(0,-dL.min[-1]),tpk)
      }
      new <- which.max(pmax.int(n2.dL.pos,n2.dL.neg))
      if (n2.dL.pos[new] > n2.dL.neg[new]) {
        active.neg[group == new] <- TRUE
      } else {
        active.pos[group == new] <- TRUE
      }
    }
  }
}

L.coop.binomial <- function(beta,X,y,lambda,group,pk,tpk) {

  ## Managing intercept
  lambda <- lambda*sqrt(tpk)
  if (length(beta) == length(group)) {
    eta <- X %*% beta
    a.pos <- group.norm(pmax.int(0, beta),tpk)
    a.neg <- group.norm(pmax.int(0,-beta),tpk)
  } else {
    eta <- cbind(1,X) %*% beta
    a.pos <- group.norm(pmax.int(0, beta[-1]),tpk)
    a.neg <- group.norm(pmax.int(0,-beta[-1]),tpk)
  }
  
  L <- - sum( y * eta - log(1 + exp(eta)) ) + sum(lambda*(a.pos + a.neg))

  return(L)
}

dL.coop.binomial <- function(beta,X,y,lambda,group,pk,tpk) {

  ## Managing intercept
  lambda <- lambda * sqrt(pk)
  if (length(beta) == length(group)) {
    eta <- X %*% beta
    nabla.f <- - t(X) %*% (y-1/(1+exp(-eta)))
    a.pos <- group.norm.rep(pmax.int(0, beta),tpk)
    a.neg <- group.norm.rep(pmax.int(0,-beta),tpk)
  } else {
    eta <- cbind(1,X) %*% beta
    nabla.f <- - ( t(cbind(1,X)) %*% (y-1/(1+exp(-eta))) )
    lambda <- c(0,lambda)
    a.pos <- c(max(0, beta[1]),group.norm.rep(pmax.int(0, beta[-1]),tpk))
    a.neg <- c(max(0,-beta[1]),group.norm.rep(pmax.int(0,-beta[-1]),tpk))
  }
    
  ## manage cases where alpha+ > 0 or alpha- < 0
  dL <- nabla.f
  neg <- beta < 0
  pos <- beta > 0
  dL[pos] <- nabla.f[pos] + lambda[pos]*pmax.int(rep(0,sum(pos)), beta[pos]) / a.pos[pos]
  dL[neg] <- nabla.f[neg] - lambda[neg]*pmax.int(rep(0,sum(neg)),-beta[neg]) / a.neg[neg]
  
  ## manage cases where beta = 0 and (alpha- = 0 and/or alpha+ = 0)
  if (length(beta) == length(group)) {
    n2.theta.pos <- group.norm.rep(pmax.int(0, nabla.f),tpk)
    n2.theta.neg <- group.norm.rep(pmax.int(0,-nabla.f),tpk)
  } else {
    n2.theta.pos <- c(max(0, nabla.f[1]),group.norm.rep(pmax.int(0, nabla.f[-1]),tpk))
    n2.theta.neg <- c(max(0,-nabla.f[1]),group.norm.rep(pmax.int(0,-nabla.f[-1]),tpk))
  }
  zero.pos  <- abs(beta) == 0 & nabla.f > 0 & a.neg == 0
  zero.neg  <- abs(beta) == 0 & nabla.f < 0 & a.pos == 0
  
  pos.to.zero  <- n2.theta.pos[zero.pos] <= lambda[zero.pos]
  neg.to.zero  <- n2.theta.neg[zero.neg] <= lambda[zero.neg]

  dL[zero.pos][!pos.to.zero] <- nabla.f[zero.pos][!pos.to.zero] * (1-lambda[zero.pos][!pos.to.zero]/n2.theta.pos[zero.pos][!pos.to.zero])
  dL[zero.neg][!neg.to.zero] <- nabla.f[zero.neg][!neg.to.zero] * (1-lambda[zero.neg][!neg.to.zero]/n2.theta.neg[zero.neg][!neg.to.zero])
  
  dL[zero.pos][pos.to.zero] <- 0
  dL[zero.neg][neg.to.zero] <- 0
  
  return(dL)
}
