scoop.env <- new.env(parent = baseenv())
assign("scoop.par", list(method = NULL, L = NULL, dL = NULL),
       envir=scoop.env)

scoop <- function(x, ...) {
  UseMethod("scoop")
}

scoop.formula <- function(formula,
                          data,
                          family       = "gaussian",
                          penalty      = "coop", 
                          intercept    = TRUE,
                          normalize    = TRUE,
                          lambda       = NULL,
                          n.lambda     = 100,
                          lambda.min   = 1e-3,
                          verbose      = FALSE,
                          eps          = 1e-5,
                          max.iter     = NULL,
                          max.grp      = NULL, ...) {

  mf <- match.call(expand.dots = FALSE)
  m  <- match(c("formula","data"), names(mf), 0L)
  mf <- mf[c(1L,m)]
  mf$drop.unused.levels <- TRUE
  mf[[1L]] <- as.name("model.frame")
  mf <- eval(mf, parent.frame())
  x  <- model.matrix(attr(mf,"terms"), data = mf, contrasts = NULL)
  y  <- model.response(mf)

  group <- attr(x, "assign")
  ## treating intercept
  if (0 %in% group) {
    group <- group[-1]
    intercept <- TRUE
    x <- x[,-1]
  } else {
    intercept <- FALSE
  }
  if (is.null(max.iter)) {
    max.iter <- 2*ncol(x)
  }
  if (is.null(max.grp)) {
    max.grp <- length(table(group))
  }
  
  out <- scoop.default(x = x,
                       y = y,
                       group        = group,
                       family       = family,
                       penalty      = penalty, 
                       intercept    = intercept,
                       normalize    = normalize,
                       lambda       = lambda,
                       n.lambda     = n.lambda,
                       lambda.min   = lambda.min,
                       verbose      = verbose,
                       eps          = eps,
                       max.iter     = max.iter,
                       max.grp      = max.grp)
  out$call  <- match.call()
  out$terms <- attr(mf,"terms")
  
  return(structure(out, class="scoop"))
}

scoop.default <- function(x,
                          y,
                          group        = 1:ncol(x),
                          family       = "gaussian",
                          penalty      = "coop", 
                          intercept    = TRUE,
                          normalize    = TRUE,
                          lambda       = NULL,
                          n.lambda     = 100,
                          lambda.min   = 1e-3,
                          verbose      = FALSE,
                          eps          = 1e-5,
                          max.iter     = 2*length(table(group)),
                          max.grp      = length(table(group)), ...) {
  
  ## Save x and y, put some colnames
  if (is.null(colnames(x))) {colnames(x) <- 1:ncol(x)}
  x.init <- as.matrix(x)
  y.init <- y

  ## model's choice and environment setting
  model <- paste(family,penalty,sep="")
  env <- get("scoop.par", envir = scoop.env)
  env$method <- switch(model,
                       "gaussiangroup" = lm.groupLasso,
                       "gaussiancoop"  = lm.coopLasso,
                       "binomialgroup" = lrm.groupLasso,
                       "binomialcoop"  = lrm.coopLasso, NULL)
  env$L      <- switch(model,
                       "gaussiangroup" = L.group.gaussian,
                       "gaussiancoop"  = L.coop.gaussian ,
                       "binomialgroup" = L.group.binomial,
                       "binomialcoop"  = L.coop.binomial , NULL)
  env$dL     <- switch(model,
                       "gaussiangroup" = dL.group.gaussian,
                       "gaussiancoop"  = dL.coop.gaussian ,
                       "binomialgroup" = dL.group.binomial,
                       "binomialcoop"  = dL.coop.binomial , NULL)
  assign("scoop.par", env, envir=scoop.env)
  if (is.null(env$method) | is.null(env$L) | is.null(env$dL)) {
    stop("wrong type of model specified")
  }

  ## ======================================================
  ## INTERCEPT - TREATMENT
  ## Gaussian case
  if (intercept & family == "gaussian") {    
    y.bar <- mean(y)
    y <- y-y.bar
    x.bar <- colMeans(x)
    x     <- scale(x,x.bar,FALSE)
    mu.start <- NULL
  }
  ## Binomial family
  if (family == "binomial") {
    mu   <- c()  
    if (intercept) {
      mu.start <- 0
      x.bar <- colMeans(x)
      x <- scale(x,x.bar,FALSE)      
    } else {
      mu.start <- NULL
    }
  }

  ## ======================================================
  ## NORMALIZATION, REORDERING
  ## normalizing the data
  if (normalize) {
    norm <- sqrt(drop(colSums(x^2)))
    x <- scale(x, FALSE, norm)
  } else {
    norm <- rep(1,ncol(x))
  } 
  ## sorting the groups and the columns of the design matrix
  o       <- order(group,decreasing=FALSE)
  x       <- x[,o]
  group   <- group[o]
  tpk     <- tabulate(factor(group))
  
  ## ======================================================
  ## GENERATE A GRID OF LAMBDA IS NONE HAS BEEN PROVIDED
  if (is.null(lambda)) {
    if (family == "binomial" & intercept) {
      xTy <- t(x) %*% (y-mean(y))
    } else {
      xTy <- t(x) %*% y
    }
    if (penalty == "coop") {
      n2p  <- group.norm(pmax.int(0,xTy),tpk)
      n2m  <- group.norm(pmin.int(0,xTy),tpk)
      nmax <- pmax.int(n2p,n2m)
    } else {
      nmax <- group.norm(xTy,tpk)
    }
    lambda.max <- max(nmax / sqrt(tpk))
    lambda <- 10^(seq(from=log10(lambda.max),to=log10(lambda.min),
                      length=n.lambda))
  }
  
  ## ========================================================
  ## LOOP OVER LAMBDA

  beta.start <- rep(0,ncol(x))
  last.beta  <- rep(-1,ncol(x))
  beta <- c()
  real.lambda <- c()

  ## quantities which won't move along the regularization path
  ## differs weither the family is gaussian or binomial
  fx <- switch(family, "gaussian" = crossprod(x,x)   , "binomial" = x)
  fy <- switch(family, "gaussian" = c(crossprod(y,x)), "binomial" = y)
    
  for (l in lambda) {
    if (verbose) cat("\nlambda =",l)
    ## CALL TO THE FUNCTION THAT DOES THE WORK
    out <- env$method(fx,fy,group,tpk,l,beta.start,mu.start,eps,max.iter)
    
    if (out$status == "converged") {
      beta.start <- out$beta
      if (family == "binomial") {mu.start <- out$mu}
    } else {
      beta.start <- rep(0,ncol(x))
      if (family == "binomial") {
        if (intercept) {
          mu.start <- 0
        } else {
          mu.start <- NULL
        }        
      }
    }

    ## When the path of solutions has stabilized, try to earn some time
    ## by avoiding irreleavant computations
    if (sqrt(sum((last.beta-out$beta)^2)) < eps & !all(last.beta == 0)) {
      missing <- n.lambda - length(real.lambda)
      beta <- rbind(beta,t(matrix(rep(out$beta,missing),ncol=missing)))
      if (family == "binomial") {
        mu <- c(mu,rep(out$mu,missing))
      }
      real.lambda <- lambda
      break
    }

    ## Move to the next lambda
    beta  <- rbind(beta,out$beta)
    real.lambda <- c(real.lambda,l)
    last.beta <- out$beta
    if (family == "binomial") {mu <- c(mu,out$mu)}

    ## When the algorithm does not converge, send a message and start over from
    ## zero for the next iterate
    if (verbose & out$status != "converged") {
      cat("\nlm.group.l1:",out$status, "with max.min|subgradient| =",
          out$max.n2dL)
      beta.start <- rep(0,ncol(x))
      if (family == "binomial") {mu.start <- 0}
    }

    ## If the number of group required by the user has been reached, then stop
    if(sum(norm2.groupwise(out$beta,group) !=0) > max.grp ) {
      break
    }
  } ## END OF THE LOOP OVER LAMBDA
  lambda <- real.lambda
  
  ## ======================================================
  ## RESCALING, REORDERING
  ## unsorting beta
  group <- group[order(o)]
  if (is.null(dim(beta))) {
    beta  <- beta[order(o)]
  } else {
    beta  <- beta[, order(o)]
  }
  x     <- x[,order(o)]
  ## re-scaling beta
  if (normalize) {
    if (length(lambda) > 1) {
      beta <- scale(as.matrix(beta), FALSE, norm)
    } else {
      beta <- beta /  norm
    }
  }
  beta <- matrix(beta,ncol=ncol(x), nrow=length(lambda))

  ## ======================================================
  ## INTERCEPT TREATMENT
  if (intercept) {
    if (family == "gaussian") {
      mu   <- y.bar -  beta %*% x.bar
    } else {
      mu <- mu - apply(t(beta) * x.bar, 2, sum)
    }
    beta <- cbind(mu,beta)    
    if (length(lambda) == 1) {
      names(beta) <- c("intercept",colnames(x))
    } else {
      colnames(beta) <- c("intercept",colnames(x))
      rownames(beta) <- round(lambda,2)
    }
  } else {    
    if (length(lambda) == 1) {
      names(beta) <- colnames(x)
    } else {
      colnames(beta) <- colnames(x)
      rownames(beta) <- round(lambda,2)
    }
  }

  ## The regular fit is used for df approximation in the linear regression case
  reg.fit <- suppressWarnings(reg.fitting(x.init, y.init, intercept, family))

  out <- list(coefficients  = beta,
              lambda        = lambda,
              group         = group,
              family        = family,
              penalty       = penalty,
              reg.fit       = reg.fit,
              x             = x.init,
              y             = y.init,
              call          = match.call())
  
  return(structure(out, class="scoop"))
}
