\name{scoop}
\alias{scoop}
\alias{scoop.default}
\alias{scoop.formula}
\title{Function to fit a penalized regression problem with coop-Lasso penalty}
\description{Fit a penalized regression problem either with a group-Lasso or
  a coop-Lasso grouping effect.}
\usage{scoop(x, ...)
\method{scoop}{formula}(formula, data,
        family     = "gaussian",
        penalty    = "coop", 
        intercept  = TRUE,
        normalize  = TRUE,
        lambda     = NULL,
        n.lambda   = 100,
        lambda.min = 1e-3,
        verbose    = FALSE,
        eps        = 1e-5,
        max.iter   = NULL,
        max.grp    = NULL, ...)

\method{scoop}{default}(x, y,
        group      = 1:ncol(x),
        family     = "gaussian",
        penalty    = "coop", 
        intercept  = TRUE,
        normalize  = TRUE,
        lambda     = NULL,
        n.lambda   = 100,
        lambda.min = 1e-3,
        verbose    = FALSE,
        eps        = 1e-5,
        max.iter   = 2*length(table(group)),
        max.grp    = length(table(group)), ...)
}

\arguments{
  \item{x}{design matrix (do NOT include intercept).}
  \item{y}{response vector.}
  \item{formula}{formula of the penalized variables. The response
    has to be on the left hand side of \code{~}.}
  \item{data}{data frame containing the variables in the model.}
  \item{group}{vector which defines how
    variables should be grouped. Components sharing the same number define a group.
  }
  \item{family}{response type. For now, either \code{"gaussian"} for linear
    regression or \code{"binomial"} for logistic regression. Default is
    \code{"gaussian"}.}
  \item{penalty}{type of grouping effect in the penalizer. For now, either
    \code{"coop"} (for cooperative-Lasso) or \code{"group"} (for
    group-Lasso).}
  \item{intercept}{logical; indicates if an intercept should be included in
    the model. Default is \code{TRUE}.}
  \item{normalize}{logical; indicates if variables should be
    normalized to have unit L2 norm before fitting. Default is
  \code{TRUE}.}
  \item{lambda}{vector of decreasing penalty levels. If \code{NULL}
    (the default), an appropriate vector will be generated with
    \code{n.lambda}  entries, starting from a level of penalty where only
    the intercept is included, and then shrinked to \code{lambda.min}.
  }
  \item{n.lambda}{integer that indicates the number of values to put
    in the \code{lambda} vector.  Ignored if \code{lambda} is provided.
  }
  \item{lambda.min}{minimal value of penalty that will be
  tried. Default is \code{1e-3}. Ignored if \code{lambda} is provided.} 
  \item{verbose}{logical; indicates verbose mode to display
  progression. Default is \code{FALSE}.} 
  \item{eps}{tolerance used to stop the optimization algorithm. Default
    is \code{1e-5}.}
  \item{max.iter}{maximum number of iterations before the optimization
  algorithm stops.
  }
  \item{max.grp}{integer that indicates the maximum number of groups to
    activate before the algorithm stops. Default is the number of level
    in \code{group} (that is, no early stop in the execution of the
  algorithm). 
  }
  \item{\dots}{only present for \code{S3} method compatibility.}
}
\value{%
  Returns an object of class \code{scoop} for which
  \code{fitted(object, ...)}, \code{predict(object, newx=NULL, ...)},
  \code{residuals(object, ...)}, \code{print(object, ...)} and
  \code{deviance(object, ...)} generic methods are available. These
  methods are undocumented since the use of the other methods is
  straightforward.  Three documented methods, \code{plot},
  \code{crossval} and \code{selection}, are also available.
  
  A \code{scoop} object is  list-like and contains the following:
  \item{coefficients}{matrix of coefficients with respect to the original input
    (even when normalized). The number of rows of \code{coefficients}
    corresponds the length of \code{lambda}.
  }
  \item{lambda}{vector of penalty levels for which the model has
  eventually been fitted.}
  \item{group}{vector which defines the grouping of the variables.}
  \item{penalty}{type of grouping effect in the penalizer: either
  \code{"coop"} or \code{"group"}. Default is \code{"coop"}.}
  \item{reg.fit}{the corresponding unpenalized fit (either OLS in the
    Gaussian case or logistic in the binomial case). In the Gaussian case,
    if \if{latex}{\eqn{n<p}}\if{html}{n<p}\if{text}{n<p}, OLS is replaced
    with ridge regression with minimal value of regularization. In the
    Binomial case, \code{reg.fit} is \code{NULL} when
    \if{latex}{\eqn{n<p}}\if{html}{n<p}\if{text}{n<p}.}
  \item{x}{design matrix saved for further use.}
  \item{y}{response vector saved for further use.}
  \item{call}{call to the \code{scoop} function.}
}
\author{Julien Chiquet}

\seealso{See also \code{\link{plot.scoop}},
  \code{\link{selection.scoop}} and \code{\link{crossval.scoop}}.}
\examples{
\dontrun{
set.seed(87)
## true parameters (including intercept)
beta <- c(2, 0.8, 1.1, 0, -1.5, 0.3, -1.0)
## two groups with three components
group <- c(1, 1, 1, 2, 2, 2)
## random design and additive noise with sd = 2
n <- 60
p <- 6 
x <- matrix(rnorm(p * n), nrow = n)
y <- cbind(1, x) \%*\% beta + rnorm(n,0,2)
## fit coop-lasso and look for BIC selection capability
coolasso <- scoop(x, y, group, lambda.min=0.1)
scoo     <- selection(coolasso)
## plot the coefficient path with the chosen model
plot(coolasso, crit=scoo$BIC)}
}
\keyword{models}
\keyword{regression}
